using System;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using tine;
using System.Windows.Forms;


namespace csSineServer
{
  [StructLayout(LayoutKind.Sequential, Pack = 1, CharSet = CharSet.Ansi)]
  public unsafe struct SineInfo
  {
    public float amplitude;
    public float frequency;
    public float noise;
    public float phase;
    public int   numberCalls;
    [MarshalAs(UnmanagedType.ByValArray, SizeConst = 64)]
      public char[] description;
  };

  public partial class Form1 : Form
  {
    private const int PRP_SINE = 1;
    private const int PRP_AMPLITUDE = 2;
    private const int PRP_FREQUENCY = 3;
    private const int PRP_PHASE = 4;
    private const int PRP_NOISE = 5;
    private const int PRP_INFO = 6;
    private const int NUM_DEVICES = 10;
    private const int NUM_VALUES = (1024*8);
    float[,] sinbuf = new float[NUM_DEVICES,NUM_VALUES];
    private SineInfo[] sineInfoTable = new SineInfo[NUM_DEVICES];
    public Int32 sineqm(string dev, string prp, TDataType dout, TDataType din, UInt16 acc,TEquipmentModule eqm)
    {
      Console.WriteLine("eqm called: "+dev+" "+prp);
      float fval = 0;
      float[] mcarray = new float[NUM_DEVICES];
      int cc;
      DTYPE dtdo = dout.GetDTYPE();
      DTYPE dtdi = din.GetDTYPE();
      int devnr = eqm.GetDeviceNumber(dev, prp); //native.GetDeviceNumberEx("SINEQM", dev, prp);
      int prpid = eqm.GetPropertyId(prp); //native.GetPropertyId("SINEQM", prp);
      switch (prpid)
      {
        case PRP_SINE:
          if ((acc & Access.CA_WRITE) != 0) return Errors.illegal_read_write;
          if ((cc = dout.PutData(sinbuf,devnr,NUM_VALUES,0)) != 0) return cc;
          sineInfoTable[devnr].numberCalls++;
          return 0;
        case PRP_AMPLITUDE:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            if ((cc = din.GetData(ref fval)) != 0) return cc;
            if (fval < 1 || fval > 1000) return Errors.out_of_range;
            sineInfoTable[devnr].amplitude = fval;
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
            //if (dout.dFormat == Formats.CF_FLTINT)
            //{
            //  for (int i = 0; i < NUM_DEVICES && i < dout.dArrayLength; i++)
            //  {
            //    ((FLTINT)dout.data)[i].fval = sineInfoTable[i].amplitude;
            //    ((FLTINT)dout.data)[i].ival = sineInfoTable[i].numberCalls;
            //  }
            //  return 0;
            //}
            for (int i = 0; i < NUM_DEVICES; i++)
              mcarray[i] = sineInfoTable[i].amplitude;
            if ((cc = dout.PutData(mcarray,NUM_DEVICES,devnr)) != 0) return cc;
          }
          return 0;
        case PRP_FREQUENCY:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            if ((cc = din.GetData(fval, 1, 0)) != 0) return cc;
            if (fval < 1 || fval > 100) return Errors.out_of_range;
            sineInfoTable[devnr].frequency = fval;
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
            for (int i = 0; i < NUM_DEVICES; i++)
              mcarray[i] = sineInfoTable[i].frequency;
            if ((cc = dout.PutData(mcarray,NUM_DEVICES,devnr)) != 0) return cc;
          }
          return 0;
        case PRP_PHASE:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            if ((cc = din.GetData(fval, 1, 0)) != 0) return cc;
            if (fval < 0 || fval > 512) return Errors.out_of_range;
            sineInfoTable[devnr].phase = fval;
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
            for (int i = 0; i < NUM_DEVICES; i++)
              mcarray[i] = sineInfoTable[i].phase;
            if ((cc = dout.PutData(mcarray,NUM_DEVICES,devnr)) != 0) return cc;
          }
          return 0;
        case PRP_NOISE:
          if (din.GetDataArrayLength() > 0)
          { /* input data => require write access */
            if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
            if ((cc = din.GetData(fval, 1, 0)) != 0) return cc;
            if (fval < 0 || fval > 100) return Errors.out_of_range;
            sineInfoTable[devnr].noise = fval;
          }
          if (dout.GetDataArrayLength() > 0)
          { /* prepare multichannel array */
            for (int i = 0; i < NUM_DEVICES; i++)
              mcarray[i] = sineInfoTable[i].noise;
            if ((cc = dout.PutData(mcarray,NUM_DEVICES,devnr)) != 0) return cc;
          }
          return 0;
        case PRP_INFO:
          if (din.GetDataArrayLength() > 0)
          { /* allow an atomic set of all relevant values this way : */
            if (din.GetDataFormat() != 512 + Formats.CF_STRUCT) return Errors.illegal_format;
            if (din.GetDataTag().CompareTo("SineInfo") != 0) return Errors.invalid_structure_tag;
            //if (strncmp(din->dTag, "SineInfo", 8)) return invalid_structure_tag;
            SineInfo sinf = new SineInfo();
            Marshal.PtrToStructure(dtdi.data, sinf);
            //sinf = (SineInfo*)(din->data.vptr);
            if (sinf.amplitude < 1 || sinf.amplitude > 1000) return Errors.out_of_range;
            if (sinf.frequency < 1 || sinf.frequency > 100) return Errors.out_of_range;
            if (sinf.phase < 0 || sinf.phase > 512) return Errors.out_of_range;
            if (sinf.noise < 0 || sinf.noise > 100) return Errors.out_of_range;
            sineInfoTable[devnr].amplitude = sinf.amplitude;
            sineInfoTable[devnr].frequency = sinf.frequency;
            sineInfoTable[devnr].phase = sinf.phase;
            sineInfoTable[devnr].noise = sinf.noise;
          }
          if (dout.GetDataArrayLength() > 0)
          {
            if (dout.GetDataFormat() != 512 + Formats.CF_STRUCT) return Errors.illegal_format;
            if (dout.GetDataTag().CompareTo("SineInfo") != 0) return Errors.invalid_structure_tag;
            if (dout.GetDataArrayLength() > NUM_DEVICES) dout.SetDataArrayLength(NUM_DEVICES);
            SineInfo[] sinf = new SineInfo[dout.GetDataArrayLength()];
            IntPtr p = dtdo.data;
            for (int i = 0; i < (int)dout.GetDataArrayLength(); i++)
            {
              sinf[i].description = new char[64];
              sinf[i].amplitude = sineInfoTable[(i + devnr) % NUM_DEVICES].amplitude;
              sinf[i].frequency = sineInfoTable[(i + devnr) % NUM_DEVICES].frequency;
              sinf[i].phase = sineInfoTable[(i + devnr) % NUM_DEVICES].phase;
              sinf[i].noise = sineInfoTable[(i + devnr) % NUM_DEVICES].noise;
              sinf[i].numberCalls = sineInfoTable[(i + devnr) % NUM_DEVICES].numberCalls;
              Array.Copy(sineInfoTable[(i + devnr) % NUM_DEVICES].description, sinf[i].description, 64);
              Marshal.StructureToPtr(sinf[i], p, true);
              p = (IntPtr)((long)p + Marshal.SizeOf(sinf[i]));
            }
          }
          return 0;
      }      
      return 0;
    }
    //public unsafe Int32 eqm1(string dev, string prp, ref DTYPE dout, ref DTYPE din, UInt16 acc, IntPtr eqmRef)
    //{
    //  Console.WriteLine("eqm called: "+dev+" "+prp);
    //  float fval = 0;
    //  float[] mcarray = new float[NUM_DEVICES];
    //  int cc;

    //  int devnr = native.GetDeviceNumberEx("SINEQM", dev, prp);
    //  int prpid = native.GetPropertyId("SINEQM", prp);
    //  switch (prpid)
    //  {
    //    case PRP_SINE:
    //      if ((acc & Access.CA_WRITE) != 0) return Errors.illegal_read_write;
    //      if ((cc = native.PutValuesFromFloatEx(ref dout, ref sinbuf[devnr,0], NUM_VALUES,0)) != 0) return cc;
    //      sineInfoTable[devnr].numberCalls++;
    //      return 0;
    //    case PRP_AMPLITUDE:
    //      if (din.dArrayLength > 0)
    //      { /* input data => require write access */
    //        if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
    //        if ((cc = native.GetValuesAsFloat(ref din, ref fval, 1)) != 0) return cc;
    //        if (fval < 1 || fval > 1000) return Errors.out_of_range;
    //        sineInfoTable[devnr].amplitude = fval;
    //      }
    //      if (dout.dArrayLength > 0)
    //      { /* prepare multichannel array */
    //        //if (dout.dFormat == Formats.CF_FLTINT)
    //        //{
    //        //  for (int i = 0; i < NUM_DEVICES && i < dout.dArrayLength; i++)
    //        //  {
    //        //    ((FLTINT)dout.data)[i].fval = sineInfoTable[i].amplitude;
    //        //    ((FLTINT)dout.data)[i].ival = sineInfoTable[i].numberCalls;
    //        //  }
    //        //  return 0;
    //        //}
    //        for (int i = 0; i < NUM_DEVICES; i++)
    //          mcarray[i] = sineInfoTable[i].amplitude;
    //        if ((cc = native.PutValuesFromFloatEx(ref dout, ref mcarray[0], NUM_DEVICES, devnr)) != 0) return cc;
    //      }
    //      return 0;
    //    case PRP_FREQUENCY:
    //      if (din.dArrayLength > 0)
    //      { /* input data => require write access */
    //        if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
    //        if ((cc = native.GetValuesAsFloat(ref din, ref fval, 1)) != 0) return cc;
    //        if (fval < 1 || fval > 100) return Errors.out_of_range;
    //        sineInfoTable[devnr].frequency = fval;
    //      }
    //      // memset(dout->data.cptr,0,1000); // cause memory overwrite !
    //      if (dout.dArrayLength > 0)
    //      { /* prepare multichannel array */
    //        for (int i = 0; i < NUM_DEVICES; i++)
    //          mcarray[i] = sineInfoTable[i].frequency;
    //        if ((cc = native.PutValuesFromFloatEx(ref dout, ref mcarray[0], NUM_DEVICES, devnr)) != 0) return cc;
    //      }
    //      return 0;
    //    case PRP_PHASE:
    //      if (din.dArrayLength > 0)
    //      { /* input data => require write access */
    //        if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
    //        if ((cc = native.GetValuesAsFloat(ref din, ref fval, 1)) != 0) return cc;
    //        if (fval < 0 || fval > 512) return Errors.out_of_range;
    //        sineInfoTable[devnr].phase = fval;
    //      }
    //      if (dout.dArrayLength > 0)
    //      { /* prepare multichannel array */
    //        for (int i = 0; i < NUM_DEVICES; i++)
    //          mcarray[i] = sineInfoTable[i].phase;
    //        if ((cc = native.PutValuesFromFloatEx(ref dout, ref mcarray[0], NUM_DEVICES, devnr)) != 0) return cc;
    //      }
    //      return 0;
    //    case PRP_NOISE:
    //      if (din.dArrayLength > 0)
    //      { /* input data => require write access */
    //        if ((acc & Access.CA_WRITE) != Access.CA_WRITE) return Errors.illegal_read_write;
    //        if ((cc = native.GetValuesAsFloat(ref din, ref fval, 1)) != 0) return cc;
    //        if (fval < 0 || fval > 100) return Errors.out_of_range;
    //        sineInfoTable[devnr].noise = fval;
    //      }
    //      if (dout.dArrayLength > 0)
    //      { /* prepare multichannel array */
    //        for (int i = 0; i < NUM_DEVICES; i++)
    //          mcarray[i] = sineInfoTable[i].noise;
    //        if ((cc = native.PutValuesFromFloatEx(ref dout, ref mcarray[0], NUM_DEVICES, devnr)) != 0) return cc;
    //      }
    //      return 0;
    //    case PRP_INFO:
    //      if (din.dArrayLength > 0)
    //      { /* allow an atomic set of all relevant values this way : */
    //        if (din.dFormat != 512+Formats.CF_STRUCT) return Errors.illegal_format;
    //        if (new String(din.dTag).CompareTo("SineInfo") != 0) return Errors.invalid_structure_tag;
    //        //if (strncmp(din->dTag, "SineInfo", 8)) return invalid_structure_tag;
    //        SineInfo sinf = new SineInfo();
    //        Marshal.PtrToStructure(din.data, sinf);
    //        //sinf = (SineInfo*)(din->data.vptr);
    //        if (sinf.amplitude < 1 || sinf.amplitude > 1000) return Errors.out_of_range;
    //        if (sinf.frequency < 1 || sinf.frequency > 100) return Errors.out_of_range;
    //        if (sinf.phase < 0 || sinf.phase > 512) return Errors.out_of_range;
    //        if (sinf.noise < 0 || sinf.noise > 100) return Errors.out_of_range;
    //        sineInfoTable[devnr].amplitude = sinf.amplitude;
    //        sineInfoTable[devnr].frequency = sinf.frequency;
    //        sineInfoTable[devnr].phase = sinf.phase;
    //        sineInfoTable[devnr].noise = sinf.noise;
    //      }
    //      if (dout.dArrayLength > 0)
    //      {
    //        if (dout.dFormat != 512 + Formats.CF_STRUCT) return Errors.illegal_format;
    //        if (new String(dout.dTag).CompareTo("SineInfo") != 0) return Errors.invalid_structure_tag;
    //        if (dout.dArrayLength > NUM_DEVICES) dout.dArrayLength = NUM_DEVICES;
    //        SineInfo[] sinf = new SineInfo[dout.dArrayLength];
    //        IntPtr p = dout.data;
    //        for (int i = 0; i < (int)dout.dArrayLength; i++)
    //        {
    //          sinf[i].description = new char[64];
    //          sinf[i].amplitude = sineInfoTable[(i + devnr) % NUM_DEVICES].amplitude;
    //          sinf[i].frequency = sineInfoTable[(i + devnr) % NUM_DEVICES].frequency;
    //          sinf[i].phase = sineInfoTable[(i + devnr) % NUM_DEVICES].phase;
    //          sinf[i].noise = sineInfoTable[(i + devnr) % NUM_DEVICES].noise;
    //          sinf[i].numberCalls = sineInfoTable[(i + devnr) % NUM_DEVICES].numberCalls;
    //          Array.Copy(sineInfoTable[(i + devnr) % NUM_DEVICES].description, sinf[i].description, 64);
    //          Marshal.StructureToPtr(sinf[i], p, true);
    //          p = (IntPtr)((long)p + Marshal.SizeOf(sinf[i]));
    //        }
    //      }
    //      return 0;
    //  }      
    //  return 0;
    //}
    void updateSine(int devnr)
    { /* used internally -> don't bother with bounds checking */
      int i, k;
      float f = sineInfoTable[devnr].frequency;
      int n = (int)sineInfoTable[devnr].noise;
      int p = (int)sineInfoTable[devnr].phase;
      float a = sineInfoTable[devnr].amplitude;
      Random rnd = new Random();
      for (i=0; i<NUM_VALUES; i++)
      {
        k = (i + p) % NUM_VALUES;
        sinbuf[devnr, k] = (float)(rnd.NextDouble())*n + (float)(Math.Sin(f * i * 6.2832 / (NUM_VALUES / 8)) * a);
      }
      return;
    }
    private TTaggedStruct sineInfo = new TTaggedStruct("SineInfo");

    void registerStructs()
    {
      ///* this must follow the order of the structure explicitly! */
      //native.AddFieldToStruct("SineInfo", 0, 1, 512 + Formats.CF_FLOAT,"amplitude");
      //native.AddFieldToStruct("SineInfo", 4, 1, 512 + Formats.CF_FLOAT, "frequency");
      //native.AddFieldToStruct("SineInfo", 8, 1, 512 + Formats.CF_FLOAT, "noise");
      //native.AddFieldToStruct("SineInfo", 12, 1, 512 + Formats.CF_FLOAT, "phase");
      //native.AddFieldToStruct("SineInfo", 16, 1, 512 + Formats.CF_INT32, "numberCalls");
      //native.AddFieldToStruct("SineInfo", 20, 64, 512 + Formats.CF_CHAR, "description");
      ///* terminate the structure definition like this! */
      //native.SealTaggedStruct("SineInfo", 84, NUM_DEVICES);

      TTaggedStruct tts = new TTaggedStruct(sineInfoTable);
 
      //sineInfo.AddField(1, Formats.CF_FLOAT, "amplitude");
      //sineInfo.AddField(1, Formats.CF_FLOAT, "frequency");
      //sineInfo.AddField(1, Formats.CF_FLOAT, "noise");
      //sineInfo.AddField(1, Formats.CF_FLOAT, "phase");
      //sineInfo.AddField(1, Formats.CF_INT32, "numberCalls");
      //sineInfo.AddField(64, Formats.CF_CHAR, "description");
      //sineInfo.Seal(Marshal.SizeOf(sineInfoTable[0]),sineInfoTable.Length);
    }
    public void sinbkg()
    //public unsafe void bkg(IntPtr eqmref)
    {
      for (int i = 0; i < NUM_DEVICES; i++) updateSine(i);
    }
    public void sinini()
    //public unsafe void fcn(IntPtr eqmref)
    {
      Console.WriteLine("function called");
      for (int i = 0; i < NUM_DEVICES; i++)
      {
        sineInfoTable[i].amplitude = 256;
        sineInfoTable[i].frequency = 1;
        sineInfoTable[i].phase = 0;
        sineInfoTable[i].noise = (float)5.0;
        String dsc = "Sine device " + i + " at your service";
        sineInfoTable[i].description = new char[64];
        Array.Copy(dsc.ToCharArray(),sineInfoTable[i].description,Math.Min(dsc.Length,64));
      }
    }
    public void sinexi()
    {
      Console.WriteLine("server stopping");
    }
    public Form1()
    {
      InitializeComponent();
    }
    //public native.TEqmFcn myFcn;
    //public native.TEqmHook myHook;
    //public native.TEqmHook myBkg;
    public TEquipmentModule myeqm;
    private void Form1_Load(object sender, EventArgs e)
    {
      //native._SystemInit(1);
      TKernel.InitializeServer();
      registerStructs();
      //myFcn = eqm1;
      //myHook = fcn;
      //myBkg = bkg;
      myeqm = new TEquipmentModule("", "SINEQM", 0, sineqm, sinini, sinbkg, 1000, sinexi);
      //native.RegisterEquipmentModuleEx("", "SINEQM", 0, myFcn, myHook, myBkg, 1000, myHook, IntPtr.Zero);
      
      TKernel.StartServices();
      //native.SystemSetCycleTimer();
    }
  }
}